"""init

Revision ID: 45223a2c0c67
Revises: 
Create Date: 2024-08-20 17:19:14.962334

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "45223a2c0c67"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # execute vector extension add
    op.execute("CREATE EXTENSION IF NOT EXISTS vector")


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
